interface Replay {
    c: number[];
    a: number;
    r: number;
    n: number[];
}
export declare class Memory {
    private maxSize;
    private batchSize;
    private memory;
    /**
     * Creates a new memory store for experience replay.
     * @constructor
     * @param {number} maxSize - The maximum size of the memory.
     * @param {number} batchSize - The size of the batch to sample for replay.
     */
    constructor(maxSize: number, batchSize: number);
    /**
     * Samples the memory for a batch of experiences.
     * @returns {Array<Replay> | null} A random sample of the memory, or null if smaller then batchSize.
     */
    sample(): Array<Replay> | null;
    /**
     * Adds a new experience to the memory.
     * @param {number[]} state - The current state of the environment.
     * @param {number} action - The action taken in the environment.
     * @param {number} reward - The reward given for the action.
     * @param {number[]} next - The next state of the environment.
     */
    add(state: number[], action: number, reward: number, next: number[]): void;
    /**
     * Returns the last experience in the memory.
     * @returns {Replay} The latest experience in the memory.
     */
    back(): Replay;
    /**
     * Resets the contents of the memory.
     */
    reset(): void;
}
export {};
