export declare class Table {
    private table;
    private actions;
    private states;
    /**
     * Create a new dictionary-based table abstraction.
     * @param {number} states - Number of rows.
     * @param {number} actions - Number of columns.
     * @constructor
     */
    constructor(states: number, actions: number);
    /**
     * Get the column values for a given state (row).
     * @param {number[]} state - The state to lookup.
     * @returns {number[]} - The column values for the given state.
     */
    get(state: number[]): number[];
    /**
     * Sets the column value for a given state (row) and action (column).
     * @param {number[]} state - The state to update.
     * @param {number} action - The action (index) to update.
     * @param {number} value - The new value to insert.
     */
    set(state: number[], action: number, value: number): void;
    /**
     * Loads a pre-existing Q-Table.
     * @param {Record<string, number[]>} table - The table to load.
     */
    load(table: Record<string, number[]>): void;
    /**
     * Returns the current state of the Q-Table.
     * @returns {Record<string, number[]>} - The table.
     */
    save(): Record<string, number[]>;
}
